/*
 * Tencent is pleased to support the open source community by making TMagicEditor available.
 *
 * Copyright (C) 2023 THL A29 Limited, a Tencent company.  All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import { JsEngine } from '@tmagic/schema';
import { isNumber } from '@tmagic/utils';

import type { default as TMagicNode } from './Node';

export const style2Obj = (style: string) => {
  if (typeof style !== 'string') {
    return style;
  }

  const obj: Record<string, any> = {};
  style.split(';').forEach((element) => {
    if (!element) {
      return;
    }

    const items = element.split(':');

    let key = items.shift();
    let value = items.join(':');

    if (!key) return;

    key = key.replace(/^\s*/, '').replace(/\s*$/, '');
    value = value.replace(/^\s*/, '').replace(/\s*$/, '');

    key = key
      .split('-')
      .map((v, i) => (i > 0 ? `${v[0].toUpperCase()}${v.substr(1)}` : v))
      .join('');

    obj[key] = value;
  });
  return obj;
};

export const fillBackgroundImage = (value: string) => {
  if (value && !/^url/.test(value) && !/^linear-gradient/.test(value)) {
    return `url(${value})`;
  }
  return value;
};

export const getTransform = (value: Record<string, string>, jsEngine: JsEngine) => {
  if (!value) return [];

  const transform = Object.entries(value).map(([transformKey, transformValue]) => {
    if (!transformValue.trim()) return '';
    if (transformKey === 'rotate' && isNumber(transformValue)) {
      transformValue = `${transformValue}deg`;
    }

    return jsEngine !== 'hippy' ? `${transformKey}(${transformValue})` : { [transformKey]: transformValue };
  });

  if (jsEngine === 'hippy') {
    return transform;
  }
  const values = transform.join(' ');
  return !values.trim() ? 'none' : values;
};

/**
 * 将dsl中的style配置转换成css，主要是将数值转成rem为单位的样式值，例如100将被转换成1rem
 * @param style Object
 * @returns Object
 */
export const transformStyle = (style: Record<string, any> | string, jsEngine: JsEngine) => {
  if (!style) {
    return {};
  }

  let styleObj: Record<string, any> = {};
  const results: Record<string, any> = {};

  if (typeof style === 'string') {
    styleObj = style2Obj(style);
  } else {
    styleObj = { ...style };
  }

  const isHippy = jsEngine === 'hippy';

  const whiteList = ['zIndex', 'opacity', 'fontWeight'];
  Object.entries(styleObj).forEach(([key, value]) => {
    if (key === 'scale' && !results.transform && isHippy) {
      results.transform = [{ scale: value }];
    } else if (key === 'backgroundImage' && !isHippy) {
      value && (results[key] = fillBackgroundImage(value));
    } else if (key === 'transform' && typeof value !== 'string') {
      results[key] = getTransform(value, jsEngine);
    } else if (!whiteList.includes(key) && value && /^[-]?[0-9]*[.]?[0-9]*$/.test(value)) {
      results[key] = isHippy ? value : `${value / 100}rem`;
    } else {
      results[key] = value;
    }
  });

  return results;
};

export const COMMON_EVENT_PREFIX = 'magic:common:events:';
export const COMMON_METHOD_PREFIX = 'magic:common:actions:';

export const CommonMethod = {
  SHOW: 'show',
  HIDE: 'hide',
  SCROLL_TO_VIEW: 'scrollIntoView',
  SCROLL_TO_TOP: 'scrollToTop',
};

export const isCommonMethod = (methodName: string) => methodName.startsWith(COMMON_METHOD_PREFIX);

export const triggerCommonMethod = (methodName: string, node: TMagicNode) => {
  const { instance } = node;

  if (!instance) return;

  switch (methodName.replace(COMMON_METHOD_PREFIX, '')) {
    case CommonMethod.SHOW:
      instance.show();
      break;

    case CommonMethod.HIDE:
      instance.hide();
      break;

    case CommonMethod.SCROLL_TO_VIEW:
      instance.$el?.scrollIntoView({ behavior: 'smooth' });
      break;

    case CommonMethod.SCROLL_TO_TOP:
      window.scrollTo({ top: 0, behavior: 'smooth' });
      break;

    default:
      break;
  }
};

export interface EventOption {
  label: string;
  value: string;
}

export const DEFAULT_EVENTS: EventOption[] = [{ label: '点击', value: `${COMMON_EVENT_PREFIX}click` }];

export const DEFAULT_METHODS: EventOption[] = [];
